from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import rasa_nlu.version

__version__ = version.__version__
